from math import atan2
import numpy as np
from scipy import linalg as sci

#Handle rand_py import
import rand_py as rand

"""
Inputs:
   H: 3x4 matrix with norm(h_i) = 1
   K: 3x4 matrix with norm(k_i) = 1
   P: 3x4 matrix
Outputs:
   theta1, theta2: 1xN angle in rads where N = num sols
"""

#Setup subproblem
def sp6_setup(H, K, P):
   #Clear all matrices to pass by ref
   H *= 0
   K *= 0
   P *= 0

   #Then pass in new values
   H += rand.rand_normal_mat(4)
   K += rand.rand_normal_mat(4)
   P += rand.rand_mat(4)

   theta1 = rand.rand_angle()
   theta2 = rand.rand_angle()
   
   #print("theta1=", theta1, "\r\ntheta2=", theta2, "\r\n")

   d1 = H[0] @ rand.rot(K[0], theta1) @ P[0] + H[1] @ rand.rot(K[1], theta2) @ P[1]
   d2 = H[2] @ rand.rot(K[2], theta1) @ P[2] + H[3] @ rand.rot(K[3], theta2) @ P[3]
   return d1, d2

def solve_2_ellipse_numeric(xm1, xn1, xm2, xn2):
   A_1 = np.transpose(xn1)@xn1
   a = A_1[0][0]
   b = 2*A_1[1][0]
   c = A_1[1][1]
   B_1 = 2*np.transpose(xm1)@xn1
   d = B_1[0][0]
   e = B_1[0][1]
   f = (np.transpose(xm1)@xm1-1)[0][0]

   A_2 = np.transpose(xn2)@xn2
   a1 = A_2[0][0]
   b1 = 2*A_2[1][0]
   c1 = A_2[1][1]
   B_2 = 2*np.transpose(xm2)@xn2
   d1 = B_2[0][0]
   e1 = B_2[0][1]
   fq = (np.transpose(xm2)@xm2-1)[0][0]

   z0 = f*a*pow(d1,2)+pow(a,2)*pow(fq,2)-d*a*d1*fq+pow(a1,2)*pow(f,2)-2*a*fq*a1*f-d*d1*a1*f+a1*pow(d,2)*fq

   z1 = e1*pow(d,2)*a1-fq*d1*a*b-2*a*fq*a1*e-f*a1*b1*d+2*d1*b1*a*f+2*e1*fq*pow(a,2)+pow(d1,2)*a*e-e1*d1*a*d-2*a*e1*a1*f-f*a1*d1*b+2*f*e*pow(a1,2)-fq*b1*a*d-e*a1*d1*d+2*fq*b*a1*d

   z2 = pow(e1,2)*pow(a,2)+2*c1*fq*pow(a,2)-e*a1*d1*b+fq*a1*pow(b,2)-e*a1*b1*d-fq*b1*a*b-2*a*e1*a1*e+2*d1*b1*a*e-c1*d1*a*d-2*a*c1*a1*f+pow(b1,2)*a*f+2*e1*b*a1*d+pow(e,2)*pow(a1,2)-c*a1*d1*d-e1*b1*a*d+2*f*c*pow(a1,2)-f*a1*b1*b+c1*pow(d,2)*a1+pow(d1,2)*a*c-e1*d1*a*b-2*a*fq*a1*c

   z3 = -2*a*a1*c*e1+e1*a1*pow(b,2)+2*c1*b*a1*d-c*a1*b1*d+pow(b1,2)*a*e-e1*b1*a*b-2*a*c1*a1*e-e*a1*b1*b-c1*b1*a*d+2*e1*c1*pow(a,2)+2*e*c*pow(a1,2)-c*a1*d1*b+2*d1*b1*a*c-c1*d1*a*b

   z4 = pow(a,2)*pow(c1,2)-2*a*c1*a1*c+pow(a1,2)*pow(c,2)-b*a*b1*c1-b*b1*a1*c+pow(b,2)*a1*c1+c*a*pow(b1,2)

   y = np.roots(np.array([z4, z3, z2, z1, z0]))

   y = np.real(y[np.real(y) == y]) #Grab only those with no imaginary parts
   
   x = -(a*fq+a*c1*y**2-a1*c*y**2+a*e1*y-a1*e*y-a1*f)/(a*b1*y+a*d1-a1*b*y-a1*d)

   return x, y #This could just return the calculations of x,y for optimization purposes

def sp6_run(H, K, P, d1, d2):
   k1Xp1 = np.cross(K[0], P[0])
   k1Xp2 = np.cross(K[1], P[1])
   k1Xp3 = np.cross(K[2], P[2])
   k1Xp4 = np.cross(K[3], P[3])

   A_1 = np.concatenate(([k1Xp1], [-np.cross(K[0], k1Xp1)]), axis = 0)
   A_2 = np.concatenate(([k1Xp2], [-np.cross(K[1], k1Xp2)]), axis = 0)
   A_3 = np.concatenate(([k1Xp3], [-np.cross(K[2], k1Xp3)]), axis = 0)
   A_4 = np.concatenate(([k1Xp4], [-np.cross(K[3], k1Xp4)]), axis = 0)
   A   = np.reshape(np.concatenate(([H[0]@A_1.T], [H[1]@A_2.T], [H[2]@A_3.T], [H[3]@A_4.T]), axis = 0), (2,4))
   
   x_min = np.linalg.lstsq(A, np.concatenate(([d1 - H[0]@np.reshape(K[0], (3,1))*K[0]@np.reshape(P[0], (3,1))-H[1]@np.reshape(K[1], (3,1))*K[1]@np.reshape(P[1], (3,1))], 
   [d2 -H[2]@np.reshape(K[2], (3,1))*K[2]@np.reshape(P[2], (3,1))-H[3]@np.reshape(K[3], (3,1))*K[3]@np.reshape(P[3], (3,1))]), axis = 0), rcond=None)

   x_min = x_min[0] #Removes residuals, matrix rank, and singularities   

   x_null = np.reshape(np.array(sci.null_space(A)), (2, 4)) #Save only the answer
   x_null_1 = np.array([x_null[0][0], x_null[0][2], x_null[1][0], x_null[1][2]])
   x_null_2 = np.array([x_null[0][1], x_null[0][3], x_null[1][1], x_null[1][3]])

   xi_1, xi_2 = solve_2_ellipse_numeric(x_min[0:2], np.reshape(x_null[0:1], (2,2)), x_min[2:4], np.reshape(x_null[1:2], (2,2)))

   theta1 = []
   theta2 = []
   for i in range(np.size(xi_1)):
      x = x_min + np.reshape(x_null_1*xi_1[i] + x_null_2*xi_2[i], (4,1))
      theta1.append(atan2(x[0], x[1]))
      theta2.append(atan2(x[2], x[3]))
   return [theta1, theta2]


def sp6_error(H, K, P, theta1, theta2, d1, d2):
   e = 0 #Setup sum variable
   for i in range(len(theta1)):
      #e_i = norm(1x2 np array)
      e_i = np.linalg.norm(np.array([np.transpose(H[0])@rand.rot(K[0], theta1[i])@P[0] +
      np.transpose(H[1])@rand.rot(K[1], theta2[i])@P[1] - d1, 
      np.transpose(H[2])@rand.rot(K[2], theta1[i])@P[2] + 
      np.transpose(H[3])@rand.rot(K[3], theta2[i])@P[3] - d2]), 2)
      e += e_i #Add to sum
   return e



#Test Code
if __name__ == "__main__":
   print("Starting arrays \r\n")
   np.set_printoptions(precision = 20)
   H = np.zeros((4, 3))
   K = np.zeros((4, 3))
   P = np.zeros((4, 3))

   #Setup problem
   d1, d2 = sp6_setup(H, K, P)
   print("d1 =", d1, "\r\nd2 =", d2, "\r\n")

   #MATLAB friendly output is below
   print(" ".join(("H=" + str(H.T) + "K=" + str(K.T) + "P=" + str(P.T)).split()).replace("] ", "];").replace("]]", "]]\n"))
   theta1, theta2 = sp6_run(H, K, P, d1, d2)

   #Printing out results
   print("Results:\r\ntheta1: {}\r\ntheta2: {}\r".format(theta1, theta2))
   print("Error:", sp6_error(H, K, P, theta1, theta2, d1, d2), "\r\n")


from math import atan2
import numpy as np

#Handle rand_py import
import rand_py as rand


#Inputs: 3x1 numpy arrays p1, p2, k
#Python does not pass by reference like C++, must manipulate without assignment
#Theta is calculated outside this function as opposed to MATLAB version
def sp1_setup(p1, p2, k, theta):
   #Clear all matrices
   p1 *= 0
   p2 *= 0
   k *= 0

   #Then fill in values
   p1 += rand.rand_vec()
   k += rand.rand_normal_vec()
   p2 += rand.rot(k, theta) @ p1 #@ is index wise multiplication
 

#Returns theta, is_LS
def sp1_run(p1, p2, k):
   #p2 *= 0
   #p2 += rand.rot(k, theta) @ p1

   KxP = np.cross(k, p1)
   A = np.vstack((KxP, -np.cross(k, KxP)))

   x = np.dot(A, p2)
   
   #theta = atan2(x[0], x[1])

   return atan2(x[0], x[1]), abs(np.linalg.norm(p1, 2) - np.linalg.norm(p2, 2)) > 1e-8 or abs(np.dot(k,p1) - np.dot(k,p2)) > 1e-8


#Inputs: 3x1 numpy arrays p1, p2, k
#Python does not pass by reference like C++, must manipulate without assignment
#Theta is calculated outside this function as opposed to MATLAB version
def sp1_setup_LS(p1, p2, k):
   #Clear all matrices
   p1 *= 0
   p2 *= 0
   k *= 0

   #Then fill in values
   p1 += rand.rand_vec()
   k += rand.rand_normal_vec()
   p2 += rand.rand_vec()


def sp1_error(p1, p2, k, theta):
   return np.linalg.norm(p2 - rand.rot(k, theta)@p1, 2)


#Testing
if __name__ == "__main__":
   print("Starting arrays \r\n")
   p1 = np.array([1., 2., 3.])
   p2 = np.array([1., 2., 3.])
   k = np.array([1., 2., 3.])
   print("Passed arrays \r\n")

   theta = rand.rand_angle()

   sp1_setup(p1, p2, k, theta)

   print("After running: ", theta, "\r\n", p1, "\r\n\n", p2, "\r\n\n", k, "\r\n\n")

   theta, is_LS =sp1_run(p1, p2, k)


from math import atan2,sqrt
import numpy as np

#Handle rand_py import
import rand_py as rand


"""
Inputs:
   p1, p2: 3x1 vector
   k1, k2: 3x1 vector w/ norm = 1
Outputs:
   theta1, theta2: 1xN angle vec in rad, N = num sols
   is_LS: 1x1 bool
"""

#Setup for non-least squares version
def sp2_setup(p1, p2, k1, k2):
   #Clear all matrices to pass by ref
   p1 *= 0
   p2 *= 0
   k1 *= 0
   k2 *= 0

   #Then pass in new values
   p1 += rand.rand_vec()
   k1 += rand.rand_normal_vec()
   k2 += rand.rand_normal_vec()

   #Calculate p2 using rand angles and above 
   #@ is index wise multiplication
   p2 += rand.rot(k1, -rand.rand_angle()) * rand.rot(k1, rand.rand_angle()) @ p1


#Setup for least-squares version
def sp2_setup_LS(p1, p2, k1, k2):
   #Clear all matrices to pass by ref
   p1 *= 0
   p2 *= 0
   k1 *= 0
   k2 *= 0

   #Then pass in new values
   p1 += rand.rand_vec()
   p2 += rand.rand_vec()
   k1 += rand.rand_normal_vec()
   k2 += rand.rand_normal_vec()


#Code called to run subproblem
def sp2_run(p1, p2, k1, k2):
   #Rescale for least-squares
   p1/=np.linalg.norm(p1, 2)
   p2/=np.linalg.norm(p2, 2)

   KxP1 = np.cross(k1, p1)
   KxP2 = np.cross(k2, p2)

   #np.block appends two arrays together
   A1 = np.block([[KxP1], [-np.cross(k1, KxP1)]])
   A2 = np.block([[KxP2], [-np.cross(k2, KxP2)]])

   radius_1_sq = np.dot(KxP1, KxP1)
   radius_2_sq = np.dot(KxP2, KxP2)
   
   k1_d_p1 = np.dot(k1,p1)
   k2_d_p2 = np.dot(k2,p2)
   k1_d_k2 = np.dot(k1,k2)

   ls_frac = 1/(1-np.dot(k1_d_k2, k1_d_k2)) #Use np.dot as ^2

   alpha_1 = np.dot(ls_frac,(k1_d_p1-np.dot(k1_d_k2,k2_d_p2)))
   alpha_2 = np.dot(ls_frac, (k2_d_p2-np.dot(k1_d_k2,k1_d_p1)))

   x_ls_1 = alpha_2*np.dot(A1, k2) / radius_1_sq
   x_ls_2 = alpha_1*np.dot(A2, k1) / radius_2_sq
   x_ls =  np.block([[x_ls_1], [x_ls_2]])

   n_sym = np.cross(k1,k2)
   pinv_A1 = A1/radius_1_sq
   pinv_A2 = A2/radius_2_sq
   A_perp_tilde = np.block([[pinv_A1], [pinv_A2]]) @ n_sym

   #See if this is not a least-squares problem
   if np.linalg.norm(x_ls[0], 2) < 1:
      xi = sqrt(1.0 - pow(np.linalg.norm(x_ls[0], 2),2)) / np.linalg.norm(A_perp_tilde[0:2],2)
      sc_1 = x_ls.flatten() + A_perp_tilde * xi
      sc_2 = x_ls.flatten() - A_perp_tilde * xi

      #Commented out for optimization purposes
      #theta1 = np.block([[atan2(sc_1[0], sc_1[1])], [atan2(sc_2[0], sc_2[1])]])
      #theta2 = np.block([[atan2(sc_1[2], sc_1[3])], [atan2(sc_2[2], sc_2[3])]])
      return np.block([[atan2(sc_1[0], sc_1[1])], [atan2(sc_2[0], sc_2[1])]]), np.block([[atan2(sc_1[2], sc_1[3])], [atan2(sc_2[2], sc_2[3])]]), False

   #Otherwise, this is least-squares
   else:
      #Same as above, it is optimal to return these values vs defining variables
      #theta1 = atan2(x_ls[0][0], x_ls[0][1])
      #theta2 = atan2(x_ls[1][0], x_ls[1][1])
      return atan2(x_ls[0][0], x_ls[0][1]), atan2(x_ls[1][0], x_ls[1][1]), True


#Error calculation code
def sp2_error(p1, p2, k1, k2, theta1,theta2):
   #If least-squares:     theta1,theta2 are floats
   #If not least-squares: theta1,theta2 are np.array
   if type(theta1) != float:
      e = 0.0 #Sum variable
      #Perform calculations on each index
      for i in range(len(theta1)):
         e_i = np.linalg.norm(np.dot(rand.rot(k2, theta2[i]), p2) - np.dot(rand.rot(k1, theta1[i]), p1), 2)
         e += e_i
   #If least-squares, then no summation needed
   else:
      e =  np.linalg.norm(np.dot(rand.rot(k2, theta2), p2) - np.dot(rand.rot(k1, theta1), p1), 2)
   return e



#Test Code
if __name__ == "__main__":
   print("Starting arrays \r\n")
   p1 = np.array([1., 2., 3.])
   p2 = np.array([1., 2., 3.])
   k1 = np.array([1., 2., 3.])
   k2 = np.array([1., 2., 3.])
   
   #Setup problem
   sp2_setup(p1, p2, k1, k2)
   t1, t2, is_LS = sp2_run(p1, p2, k1, k2) #Save values

   #Printing out results
   print("Results:\r\ntheta1:\n{}\r\ntheta2:\n{}\nis_LS:\n{}\r\n".format(t1, t2, is_LS))
   print("Error:", sp2_error(p1, p2, k1, k2, t1, t2), "\r\n")


from math import atan2,sqrt
import numpy as np

#Handle rand_py import
import rand_py as rand

"""
Inputs:
   p1, p2: 3x1 vector
   k: 3x1 vector w/ norm = 1
   d: 1x1 dependent on is_LS
Outputs:
   theta1: 1xN angle vec in rad, N = num sols
   is_LS: 1x1 bool

This problem will be ill-posed if p1 or p2 is parallel to k
"""

#Setup for non-least squares version
def sp3_setup(p1, p2, k):
   #Clear all matrices to pass by ref
   p1 *= 0
   p2 *= 0
   k *= 0

   #Then pass in new values
   p1 += rand.rand_vec()
   p2 += rand.rand_vec()
   k += rand.rand_normal_vec()
   
   theta = rand.rand_angle()
   #print("theta=", theta, "\r\n") #If seeing theta is desired, uncomment this.

   #Return d
   return np.linalg.norm(p2 - rand.rot(k, theta) @ p1, 2)


#Setup for least squares version
def sp3_setup_LS(p1, p2, k):
   #Clear all matrices to pass by ref
   p1 *= 0
   p2 *= 0
   k *= 0

   #Then pass in new values
   p1 += rand.rand_vec()
   p2 += rand.rand_vec()
   k += rand.rand_normal_vec()

   #Return d
   return rand.r.random() #Random num between 0-1
   

#Run the problem and return theta
def sp3_run(p1, p2, k, d):
   KxP = np.cross(k, p1)
   #np.concatenate stacks two arrays here: [arr1; arr2]
   A1 = np.concatenate(([KxP], [-np.cross(k, KxP)]), axis = 0)
   A = -2*p2@A1.T

   norm_A_sq = np.dot(A,A)
   norm_A = sqrt(norm_A_sq)
   b = pow(d,2) - pow(np.linalg.norm(p2 - np.reshape(k, (3,1)) @ np.reshape(k, (1,3)) @ p1, 2), 2) - pow(np.linalg.norm(KxP, 2), 2)

   x_ls = A1@(-2*p2*b/norm_A_sq)

   #Check to see if this is least squares version
   if x_ls @ x_ls > 1:
      #theta = atan2(x_ls[0], x_ls[1])
      return atan2(x_ls[0], x_ls[1]), True
   
   #Otherwise not a least squares
   else:
      xi = sqrt(1-pow(b, 2)/norm_A_sq)

      A_perp_tilde = np.block([A[1], -A[0]])
      A_perp = A_perp_tilde/norm_A

      sc_1 = x_ls + xi * A_perp
      sc_2 = x_ls - xi * A_perp

      #theta = [atan2(sc_1[0], sc_1[1]), atan2(sc_2[0], sc_2[1])]
      return np.array([atan2(sc_1[0], sc_1[1]), atan2(sc_2[0], sc_2[1])]), False


#Calculate error
def sp3_error(p1, p2, k, d, theta):
   #If least-squares:     theta is a float
   #If not least-squares: theta is a np.array
   if type(theta) != float:
      e = 0.0 #Sum variable
      #Perform calculations on each index
      for i in range(len(theta)):
         e_i = abs(np.linalg.norm(p2 - rand.rot(k, theta[i]) @ p1, 2) - d)
         e += e_i
   #If least-squares, then no summation needed
   else:
      e = abs(np.linalg.norm(p2 - rand.rot(k, theta) @ p1, 2) - d)
   return e



#Test Code
if __name__ == "__main__":
   print("Starting arrays \r\n")
   p1 = np.array([1., 2., 3.])
   p2 = np.array([1., 2., 3.])
   k = np.array([1., 2., 3.])

   #Setup problem
   print("After setup: \r\n")
   d = sp3_setup_LS(p1, p2, k)
   np.set_printoptions(precision = 20)
   print("p1 = {}\r\np2 = {}\r\nk  = {}\r\nd  = {}\r\n".format(p1, p2, k, d))

   theta, is_LS = sp3_run(p1, p2, k, d) #Save values

   #Printing out results
   print("Results:\r\ntheta: {}\r\nis_LS: {}\r".format(theta, is_LS))
   print("Error:", sp3_error(p1, p2, k, d, theta), "\r\n")